/* Net.c        -- Network code for Palm LiveJournal
 * $Id: Net.c,v 1.2 2001/03/30 23:52:52 kvance Exp $
 * Copyright (C) 2000-2001 Kev Vance <kvance@tekktonik.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <PalmOS.h>
#include <System/NetMgr.h>
#include <System/DateTime.h>
#include <INet/INetMgr.h>
#include <CTP.h>

#include "PalmLJ.h"

UInt16 AppNetRefnum;
Int32 AppNetTimeout;

void UrlCat(Char * x, Char * y)
{
    int i, j, hex;
    Err err;

    j = StrLen(x);		// The end of the dest. string

    for (i = 0; i < StrLen(y); i++) {
	if (MemPtrSize(x) < j + 4) {
	    /* Out of memory, request more */
	    err = MemPtrResize(x, MemPtrSize(x) + 100);
	    if (err)
		/* Can't get more memory, abort! */
		break;
	}
	if ((y[i] >= 'a' && y[i] <= 'z') ||	// a-z
	     (y[i] >= 'A' && y[i] <= 'Z') ||	// A-Z
	     (y[i] >= '0' && y[i] <= '9')) {	// 0-9

	    x[j++] = y[i];
	    x[j] = '\0';
	} else if (y[i] == ' ') {
	    x[j++] = '+';
	    x[j] = '\0';
	} else {
	    x[j++] = '%';
	    /* Left hex digit */
	    hex = y[i] / 16;
	    if (hex < 10)
		hex += '0';
	    else
		hex += 'A' - 10;
	    x[j++] = hex;
	    /* Right hex digit */
	    hex = y[i] % 16;
	    if (hex < 10)
		hex += '0';
	    else
		hex += 'A' - 10;
	    x[j++] = hex;
	    x[j] = '\0';
	}
    }
}

void MemCat(Char * x, Char * y)
{
    int i, j;
    Err err;

    j = StrLen(x);		// The end of the dest. string

    for (i = 0; i < StrLen(y); i++) {
	if (MemPtrSize(x) < j + 2) {
	    /* Out of memory, request more */
	    err = MemPtrResize(x, j + 100);
	    if (err)
		/* Can't get more memory, abort! */
		break;
	}
	x[j++] = y[i];
	x[j] = '\0';
    }
}

int Post(Char * user, Char * usejournal, Char * passwd, Char * subject,
	 Char * entry, UInt16 mood, Char * custommood, UInt16 ct)
{
    NetSocketRef sock;
    NetHostInfoPtr nhinfo;
    NetHostInfoBufType nh;
    NetSocketAddrINType sain;

    Err ifErrs;
    Int16 err;

    Char *dest;
    Char *buffer;

    UInt32 seconds;
    DateTimeType date;

    /* INetLib stuff */
    UInt16 config;
    MemHandle inetHnd, inetSock;
    UInt32 compress, readbytes;

    AppNetTimeout = SysTicksPerSecond() * 15;

    /* Load the net library  & create the socket */
    switch (ct) {
    case NetLib:
	if (SysLibFind("Net.lib", &AppNetRefnum))
	    return 0;
	err = NetLibOpen(AppNetRefnum, &ifErrs);
	if (err || ifErrs)
	    return 0;
	sock = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET,
				netSocketTypeStream, 0, AppNetTimeout,
				&ifErrs);
	if (ifErrs) {
	    NetLibClose(AppNetRefnum, false);
	    return 0;
	}
	break;
    case INetLib:
	if (SysLibFind("INet.lib", &AppNetRefnum))
	    return 0;
	if (INetLibConfigIndexFromName(AppNetRefnum,
			       (INetConfigNamePtr) inetCfgNameCTPDefault,
				       &config))
	    return 0;
	if (INetLibOpen(AppNetRefnum, config, 0, NULL, 0, &inetHnd))
	    return 0;
	compress = ctpConvNone;
	if (INetLibSettingSet(AppNetRefnum, inetHnd,
			      inetSettingConvAlgorithm, &compress,
			      sizeof(compress)))
	    return 0;
	if (INetLibSockOpen(AppNetRefnum, inetHnd, inetSchemeHTTP,
			    &inetSock)) {
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	break;
    }

    /* Lookup/Connect to the LJ server */
    switch (ct) {
    case NetLib:
	nhinfo = NetLibGetHostByName(AppNetRefnum,
			       "www.livejournal.com", &nh, AppNetTimeout,
				     &ifErrs);
	if (ifErrs) {
	    NetLibClose(AppNetRefnum, false);
	    return 0;
	}
	/* Fill in the sain structure */
	MemSet(&sain, sizeof(sain), 0);
	sain.addr = nh.address[0];
	sain.family = netSocketAddrINET;
	sain.port = NetHToNS(80);

	/* Connect to the LJ server */
	err = NetLibSocketConnect(AppNetRefnum, sock,
			      (NetSocketAddrType *) & sain, sizeof(sain),
				  AppNetTimeout, &ifErrs);
	if (err || ifErrs) {
	    NetLibClose(AppNetRefnum, false);
	    return 0;
	}
	break;
    case INetLib:
	if (INetLibSockConnect(AppNetRefnum, inetSock, "www.livejournal.com",
			       80, AppNetTimeout)) {
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	break;
    }

    if (ct == NetLib) {
	/* Send HTTP headers */
	NetLibSend(AppNetRefnum, sock, "POST /cgi-bin/log.cgi HTTP/1.0\nHost: www.livejournal.com\nContent-type: multipart/form-data\nContent-length: ", 107, 0, 0, 0, AppNetTimeout, &ifErrs);
    }
    /* Allocate some memory */
    dest = MemPtrNew(5000);
    buffer = MemPtrNew(1024);
    if (dest == NULL) {
	if (ct == NetLib) {
	    NetLibSocketClose(AppNetRefnum, sock, AppNetTimeout, &ifErrs);
	    NetLibClose(AppNetRefnum, false);
	} else if (ct == INetLib) {
	    INetLibSockClose(AppNetRefnum, inetSock);
	    INetLibClose(AppNetRefnum, inetHnd);
	}
	return 0;
    }
    /* Put the date in a string */
    seconds = TimGetSeconds();
    TimSecondsToDateTime(seconds, &date);

    /* Build the HTTP POST request */
    StrCopy(dest, "mode=postevent&user=");
    UrlCat(dest, user);
    MemCat(dest, "&password=");
    UrlCat(dest, passwd);
    MemCat(dest, "&event=");
    UrlCat(dest, entry);
    MemCat(dest, "&lineendings=unix&subject=");
    if (subject != NULL)
	UrlCat(dest, subject);
    MemCat(dest, "&year=");
    StrIToA(buffer, date.year);
    MemCat(dest, buffer);
    MemCat(dest, "&mon=");
    StrIToA(buffer, date.month);
    MemCat(dest, buffer);
    MemCat(dest, "&day=");
    StrIToA(buffer, date.day);
    MemCat(dest, buffer);
    MemCat(dest, "&hour=");
    StrIToA(buffer, date.hour);
    MemCat(dest, buffer);
    MemCat(dest, "&min=");
    StrIToA(buffer, date.minute);
    MemCat(dest, buffer);
    if (mood == 1) {
	MemCat(dest, "&prop_current_mood=");
	UrlCat(dest, custommood);
    } else if (mood != 0) {
	MemCat(dest, "&prop_current_moodid=");
	StrIToA(buffer, mood - 2);
	MemCat(dest, buffer);
    }
    if (usejournal != NULL) {
	MemCat(dest, "&usejournal=");
	UrlCat(dest, usejournal);
    }
    MemCat(dest, "\n");

    switch (ct) {
    case NetLib:
	StrIToA(buffer, StrLen(dest) - 1);
	NetLibSend(AppNetRefnum, sock, buffer, StrLen(buffer), 0, 0, 0,
		   AppNetTimeout, &ifErrs);
	NetLibSend(AppNetRefnum, sock, "\n\n", 2, 0, 0, 0, AppNetTimeout,
		   &ifErrs);
	NetLibSend(AppNetRefnum, sock, dest, StrLen(dest), 0, 0, 0,
		   AppNetTimeout, &ifErrs);
	break;
    case INetLib:
	if (INetLibSockHTTPReqCreate(AppNetRefnum, inetSock, "POST",
			    "http://www.livejournal.com/cgi-bin/log.cgi",
				     NULL)) {
	    INetLibSockClose(AppNetRefnum, inetSock);
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	if (INetLibSockHTTPReqSend(AppNetRefnum, inetSock, dest,
				   StrLen(dest), AppNetTimeout)) {
	    INetLibSockClose(AppNetRefnum, inetSock);
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	/* We have to read everything it sends back :P */
	do {
	    if (INetLibSockRead(AppNetRefnum, inetSock, buffer,
				1024, &readbytes, AppNetTimeout))
		break;
	} while (readbytes != 0);
	break;
    }

    /* Close the socket/library */
    switch (ct) {
    case NetLib:
	NetLibSocketClose(AppNetRefnum, sock, AppNetTimeout, &ifErrs);
	NetLibClose(AppNetRefnum, false);
	break;
    case INetLib:
	INetLibSockClose(AppNetRefnum, inetSock);
	INetLibClose(AppNetRefnum, inetHnd);
	break;
    }

    MemPtrFree(dest);
    MemPtrFree(buffer);

    return -1;
}

int Login(Char * user, Char * passwd, LJPrefs * prefs, Boolean moods,
	  Boolean new, UInt16 ct)
{
    NetSocketRef sock;
    NetHostInfoPtr nhinfo;
    NetHostInfoBufType nh;
    NetSocketAddrINType sain;

    Err ifErrs;
    Int16 err;

    Char *dest;
    Char *dp;
    Char *buffer;
    Int16 i, acount;

    /* INetLib stuff */
    UInt16 config;
    MemHandle inetHnd, inetSock;
    UInt32 compress, readbytes;

    int searchstate = 0;
    Boolean success;

    UserRecord *r = NULL;
    AccessRecord *a = NULL, *b, *c;

#define R_NAME		1
#define R_ACCESS	2

    AppNetTimeout = SysTicksPerSecond() * 15;

    /* Load the net library  & create the socket */
    switch (ct) {
    case NetLib:
	if (SysLibFind("Net.lib", &AppNetRefnum))
	    return 0;
	err = NetLibOpen(AppNetRefnum, &ifErrs);
	if (err || ifErrs)
	    return 0;

	sock = NetLibSocketOpen(AppNetRefnum, netSocketAddrINET,
				netSocketTypeStream, 0, AppNetTimeout,
				&ifErrs);
	if (ifErrs) {
	    NetLibClose(AppNetRefnum, false);
	    return 0;
	}
	break;
    case INetLib:
	if (SysLibFind("INet.lib", &AppNetRefnum))
	    return 0;
	if (INetLibConfigIndexFromName(AppNetRefnum,
			       (INetConfigNamePtr) inetCfgNameCTPDefault,
				       &config))
	    return 0;
	if (INetLibOpen(AppNetRefnum, config, 0, NULL, 0, &inetHnd))
	    return 0;
	compress = ctpConvNone;
	if (INetLibSettingSet(AppNetRefnum, inetHnd,
			      inetSettingConvAlgorithm, &compress,
			      sizeof(compress)))
	    return 0;
	if (INetLibSockOpen(AppNetRefnum, inetHnd, inetSchemeHTTP,
			    &inetSock)) {
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	break;
    }

    /* Lookup/Connect to the LJ server */
    switch (ct) {
    case NetLib:
	nhinfo = NetLibGetHostByName(AppNetRefnum, "www.livejournal.com", &nh,
				     AppNetTimeout, &ifErrs);
	if (ifErrs) {
	    NetLibClose(AppNetRefnum, false);
	    return 0;
	}
	/* Fill in the sain structure */
	MemSet(&sain, sizeof(sain), 0);
	sain.addr = nh.address[0];
	sain.family = netSocketAddrINET;
	sain.port = NetHToNS(80);

	/* Connect to the LJ server */
	err = NetLibSocketConnect(AppNetRefnum, sock,
			      (NetSocketAddrType *) & sain, sizeof(sain),
				  AppNetTimeout, &ifErrs);
	if (err || ifErrs) {
	    NetLibClose(AppNetRefnum, false);
	    return 0;
	}
	break;
    case INetLib:
	if (INetLibSockConnect(AppNetRefnum, inetSock, "www.livejournal.com",
			       80, AppNetTimeout)) {
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	break;
    }

    if (ct == NetLib) {
	/* Send HTTP headers */
	NetLibSend(AppNetRefnum, sock, "POST /cgi-bin/log.cgi HTTP/1.0\nHost: www.livejournal.com\nContent-type: multipart/form-data\nContent-length: ", 107, 0, 0, 0, AppNetTimeout, &ifErrs);
    }
    /* Allocate some memory */
    dest = MemPtrNew(5000);
    buffer = MemPtrNew(256);
    if (dest == NULL) {
	NetLibSocketClose(AppNetRefnum, sock, AppNetTimeout, &ifErrs);
	NetLibClose(AppNetRefnum, false);
	return 0;
    }
    /* Build the HTTP POST request */
    StrCopy(dest, "mode=login&user=");
    UrlCat(dest, user);
    MemCat(dest, "&password=");
    UrlCat(dest, passwd);
    MemCat(dest, "&clientversion=PalmOS-PalmLJ/1.1.0");
    if (moods == true) {
	MemCat(dest, "&getmoods=");
	StrIToA(buffer, prefs->moodCount);
	MemCat(dest, buffer);
    }
    MemCat(dest, "\n");

    switch (ct) {
    case NetLib:
	StrIToA(buffer, StrLen(dest) - 1);
	NetLibSend(AppNetRefnum, sock, buffer, StrLen(buffer), 0, 0, 0,
		   AppNetTimeout, &ifErrs);
	NetLibSend(AppNetRefnum, sock, "\n\n", 2, 0, 0, 0, AppNetTimeout,
		   &ifErrs);
	NetLibSend(AppNetRefnum, sock, dest, StrLen(dest), 0, 0, 0,
		   AppNetTimeout, &ifErrs);
	break;
    case INetLib:
	if (INetLibSockHTTPReqCreate(AppNetRefnum, inetSock, "POST",
			    "http://www.livejournal.com/cgi-bin/log.cgi",
				     NULL)) {
	    INetLibSockClose(AppNetRefnum, inetSock);
	    INetLibClose(AppNetRefnum, inetHnd);
	    return 0;
	}
	if (INetLibSockHTTPReqSend(AppNetRefnum, inetSock, dest,
				   StrLen(dest), AppNetTimeout)) {
	    INetLibSockClose(AppNetRefnum, inetSock);
	    INetLibClose(AppNetRefnum, inetSock);
	    return 0;
	}
    }

    if (ct == NetLib)
	i = NetLibReceive(AppNetRefnum, sock, dest, 5000, 0, NULL, NULL,
			  AppNetTimeout, &ifErrs);
    else {
	INetLibSockRead(AppNetRefnum, inetSock, dest, 5000, &readbytes,
			AppNetTimeout);
	i = readbytes;
    }

    dest[i] = '\0';

    /* Close the socket/library */
    switch (ct) {
    case NetLib:
	NetLibSocketClose(AppNetRefnum, sock, AppNetTimeout, &ifErrs);
	NetLibClose(AppNetRefnum, false);
	/* Jump to relevant data, too */
	dp = StrStr(dest, "\r\n\r\n");
	if (dp == NULL)
	    /* Invalid response, where's my newlines? */
	    return 0;
	dp += 4;
	break;
    case INetLib:
	INetLibSockClose(AppNetRefnum, inetSock);
	INetLibClose(AppNetRefnum, inetHnd);
	dp = dest;
	break;
    }


    success = false;
    acount = 0;
    /* Add the user */
    for (i = 0; i < 256; i++) {
	if (dp[i] == '\n') {
	    /* end of line */
	    dp[i] = '\0';
	    if (searchstate == R_NAME) {
		/* We got the name!  */
		if (new == false) {
		    /* Set r to existing user */
		    for (i = 0; i < prefs->userCount; i++) {
			r = GetUserRecord(prefs->userRecord, i);
			if (!StrCompare(r->user, user))
			    break;
		    }
		    MemMove(r->name, dp, StrLen(dp));
		    success = true;
		    /* Write access records */
		    r->accessCount = acount;
		    r->accessRecord = a;
		    break;
		} else if (prefs->userCount == 0) {
		    prefs->userRecord =
			MemPtrNew(sizeof(UserRecord));
		    r = prefs->userRecord;
		} else {
		    r = GetUserRecord(prefs->userRecord, prefs->userCount - 1);
		    r->next = MemPtrNew(sizeof(UserRecord));
		    r = r->next;
		}
		prefs->userCount++;
		r->accessCount = 0;
		r->accessRecord = NULL;
		r->next = NULL;
		MemMove(r->user, user, StrLen(user) + 1);
		MemMove(r->passwd, passwd, StrLen(passwd) + 1);
		MemMove(r->name, dp, StrLen(dp) + 1);
		dp[i++] = '\n';
		success = true;
		if (a != NULL) {
		    /* We also have some access records */
		    r->accessRecord = a;
		    r->accessCount = acount;
		}
		break;
	    } else if (searchstate == R_ACCESS) {
		/* Get access record */
		acount++;
		if (a == NULL) {
		    a = MemPtrNew(sizeof(AccessRecord));
		    MemMove(a->user, dp, StrLen(dp) + 1);
		    a->next = NULL;
		} else {
		    b = GetAccessRecord(a, acount - 2);
		    c = MemPtrNew(sizeof(AccessRecord));
		    MemMove(c->user, dp, StrLen(dp) + 1);
		    b->next = c;
		}
		if (r != NULL) {
		    /* If we have a name, write to it */
		    r->accessRecord = a;
		    r->accessCount = acount;
		}
		searchstate = 0;
		dp[i++] = '\n';
		dp += i;
		i = 0;
	    } else if (!StrCompare(dp, "name")) {
		/* We found it! */
		dp[i++] = '\n';
		dp += i;
		i = 0;
		if (r == NULL)
		    /* Only get the name once */
		    searchstate = R_NAME;
	    } else if (!StrNCompare(dp, "access", 6)) {
		if (StrCompare(dp, "access_count"))
		    /* Count?  We don't need no steenkin' count */
		    searchstate = R_ACCESS;
		dp[i++] = '\n';
		dp += i;
		i = 0;
	    } else {
		dp[i++] = '\n';
		dp += i;
		i = 0;
	    }
	}
    }

    MemPtrFree(dest);
    MemPtrFree(buffer);

    if (success == false)
	return 0;
    return -1;
}
