/* PalmLJ.c     -- Main code for Palm LiveJournal
 * $Id: PalmLJ.c,v 1.2 2001/03/30 23:52:52 kvance Exp $
 * Copyright (C) 2000-2001 Kev Vance <kvance@tekktonik.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <PalmOS.h>
#include "PalmLJRsc.h"
#include "PalmLJ.h"

#define idCreator 'PaLJ'

LJPrefs *prefs;
static Int16 *usermap = NULL;
static Int16 *accessmap = NULL;

UserRecord *DelUserRecord(UserRecord * r)
{
    AccessRecord *a;
    UserRecord *s;
    int i;

    for (i = r->accessCount; i != 0; i--) {
	a = GetAccessRecord(r->accessRecord, i - 1);
	MemPtrFree(a);
    }
    s = r->next;
    MemPtrFree(r);
    return s;
}

UserRecord *GetUserRecord(UserRecord * first, UInt16 record)
{
    int i;
    UserRecord *r = first;

    for (i = 0; i != record; i++) {
	if (r->next == NULL)
	    return NULL;
	r = r->next;
    }

    return r;
}

AccessRecord *GetAccessRecord(AccessRecord * first, UInt16 record)
{
    int i;
    AccessRecord *a = first;

    for (i = 0; i != record; i++) {
	if (a->next == NULL)
	    return NULL;
	a = a->next;
    }

    return a;
}

static UInt16 GetPrefs(LJPrefs * prefs)
{
    Char *rawprefs;
    Int16 i, j, k;
    UInt16 size = 0;
    UserRecord *r, *s;
    AccessRecord *a, *b;

    i = PrefGetAppPreferences(idCreator, 0, NULL, &size, true);
    if (i == noPreferenceFound) {
	/* Create the blank prefs */
	prefs->version = 2;
	prefs->userCount = 0;
	prefs->moodCount = 0;
	prefs->lastUser = 0;
	prefs->lastAccess = 0;
	prefs->connectType = Undecided;
	prefs->userRecord = NULL;
	prefs->moodRecord = NULL;
	rawprefs = MemPtrNew(12);
	MemMove(&rawprefs[0], &prefs->version, 2);
	MemMove(&rawprefs[2], &prefs->userCount, 2);
	MemMove(&rawprefs[4], &prefs->moodCount, 2);
	MemMove(&rawprefs[6], &prefs->lastUser, 2);
	MemMove(&rawprefs[8], &prefs->lastAccess, 2);
	MemMove(&rawprefs[10], &prefs->connectType, 2);
	PrefSetAppPreferences(idCreator, 0, 1, rawprefs, 12, true);
	MemPtrFree(rawprefs);
    } else {
	/* Load and fill */
	if (size < 10)
	    /* That's not right */
	    return 0;
	rawprefs = MemPtrNew(size);
	PrefGetAppPreferences(idCreator, 0, rawprefs, &size, true);
	MemMove(&prefs->version, &rawprefs[0], 2);
	if (prefs->version > 2)
	    /* We're in over our heads! */
	    return 0;
	MemMove(&prefs->userCount, &rawprefs[2], 2);
	MemMove(&prefs->moodCount, &rawprefs[4], 2);
	MemMove(&prefs->lastUser, &rawprefs[6], 2);
	MemMove(&prefs->lastAccess, &rawprefs[8], 2);
	/* v1 has no connectType
	   also, user records are at a different position */
	if (prefs->version == 1) {
	    prefs->connectType = Undecided;
	    j = 8;
	} else {
	    MemMove(&prefs->connectType, &rawprefs[10], 2);
	    j = 12;
	}
	prefs->userRecord = NULL;
	prefs->moodRecord = NULL;
	for (i = 0; i < prefs->userCount; i++) {
	    /* Load base user record */
	    r = MemPtrNew(sizeof(UserRecord));
	    r->accessRecord = NULL;
	    r->next = NULL;
	    MemMove(&r->user, &rawprefs[j], 64);
	    j += 64;
	    MemMove(&r->passwd, &rawprefs[j], 64);
	    j += 64;
	    MemMove(&r->name, &rawprefs[j], 128);
	    j += 128;
	    MemMove(&r->accessCount, &rawprefs[j], 2);
	    j += 2;
	    /* Load access records */
	    for (k = 0; k < r->accessCount; k++) {
		a = MemPtrNew(sizeof(AccessRecord));
		a->next = NULL;
		MemMove(&a->user, &rawprefs[j], 64);
		j += 64;
		if (r->accessRecord == NULL) {
		    r->accessRecord = a;
		} else {
		    b = GetAccessRecord(r->accessRecord, k - 1);
		    b->next = a;
		}
	    }
	    /* Store it */
	    if (prefs->userRecord == NULL) {
		prefs->userRecord = r;
	    } else {
		s = GetUserRecord(prefs->userRecord, i - 1);
		s->next = r;
	    }
	}
	MemPtrFree(rawprefs);
    }
    return -1;
}

static UInt16 SavePrefs(LJPrefs * prefs)
{
    Char *rawprefs;
    Int16 i, j, k;
    UserRecord *r;
    AccessRecord *a;
    Int16 size;

//    header  |        userRecords
    size = 12 + (64 + 64 + 128 + 2) * prefs->userCount;
    for (i = 0; i < prefs->userCount; i++) {
	r = GetUserRecord(prefs->userRecord, i);
	size += 64 * r->accessCount;
    }
    rawprefs = MemPtrNew(size);
    MemMove(&rawprefs[0], &prefs->version, 2);
    MemMove(&rawprefs[2], &prefs->userCount, 2);
    MemMove(&rawprefs[4], &prefs->moodCount, 2);
    MemMove(&rawprefs[6], &prefs->lastUser, 2);
    MemMove(&rawprefs[8], &prefs->lastAccess, 2);
    MemMove(&rawprefs[10], &prefs->connectType, 2);
    for (i = 0, j = 12; i < prefs->userCount; i++) {
	r = GetUserRecord(prefs->userRecord, i);
	MemMove(&rawprefs[j], &r->user, 64);
	j += 64;
	MemMove(&rawprefs[j], &r->passwd, 64);
	j += 64;
	MemMove(&rawprefs[j], &r->name, 128);
	j += 128;
	MemMove(&rawprefs[j], &r->accessCount, 2);
	j += 2;
	for (k = 0; k < r->accessCount; k++) {
	    a = GetAccessRecord(r->accessRecord, k);
	    MemMove(&rawprefs[j], &a->user, 64);
	    j += 64;
	}
    }
    PrefSetAppPreferences(idCreator, 0, 1, rawprefs, size, true);
    MemPtrFree(rawprefs);

    return -1;
}

/* Yoink!  Thanks, O'Reilly */
static void UpdateScrollBar(void)
{
    FormPtr frm = FrmGetActiveForm();
    ScrollBarPtr scroll;
    FieldPtr field;
    UInt16 currentPosition;
    UInt16 textHeight;
    UInt16 fieldHeight;
    UInt16 maxValue;

    field = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, EntryField));
    FldGetScrollValues(field, &currentPosition, &textHeight, &fieldHeight);

    if (textHeight > fieldHeight)
	maxValue = textHeight - fieldHeight;
    else if (currentPosition)
	maxValue = currentPosition;
    else
	maxValue = 0;

    scroll = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, EntryScroll));

    SclSetScrollBar(scroll, currentPosition, 0, maxValue, fieldHeight - 1);
}
static void ScrollLines(int numLinesToScroll, Boolean redraw)
{
    FormPtr frm = FrmGetActiveForm();
    FieldPtr field;

    field = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, EntryField));
    if (numLinesToScroll < 0)
	FldScrollField(field, -numLinesToScroll, winUp);
    else
	FldScrollField(field, numLinesToScroll, winDown);

    if ((FldGetNumberOfBlankLines(field) && numLinesToScroll < 0) ||
	redraw)
	UpdateScrollBar();
}

static Boolean MainFormHandleEvent(EventPtr e)
{
    Boolean handled = false;
    FieldPtr sfield, efield, mfield;
    ListPtr mlist;
    FormPtr frm;
    ControlPtr trigger;

    UserRecord *r;
    AccessRecord *a;

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	UpdateScrollBar();
	efield = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, EntryField));
	efield->attr.hasScrollBar = true;	// PilRC won't let us set that

	FrmDrawForm(frm);
	if (!GetPrefs(prefs))
	    FrmAlert(PrefsInvalidAlert);
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	switch (e->data.menu.itemID) {
	case ConnectionItem:
	    FrmPopupForm(ConnectionForm);
	    break;
	case UserItem:
	    FrmPopupForm(UserForm);
	    break;
	case MoodsItem:
	    if (prefs->lastUser == 0) {
		FrmAlert(NoUserAlert);
	    }
	    break;
	case AboutItem:
	    FrmPopupForm(AboutForm);
	    break;
	}

	handled = true;
	break;

    case popSelectEvent:
	frm = FrmGetFormPtr(MainForm);
	mfield = FrmGetObjectPtr(frm,
				 FrmGetObjectIndex(frm, MoodField));
	mlist = FrmGetObjectPtr(frm,
				FrmGetObjectIndex(frm, MoodList));
	if (LstGetSelection(mlist) == 1) {
	    /* Custom mood */
	    FldSetUsable(mfield, true);
	    FldDrawField(mfield);
	    LstSetSelection(mlist, 1);
	} else {
	    FldSetText(mfield, NULL, 0, 0);
	    FldEraseField(mfield);
	    FldSetUsable(mfield, false);
	    FrmUpdateForm(MainForm, frmRedrawUpdateCode);
	}
	handled = false;
	break;

	/* ScrollBar events */
    case fldChangedEvent:
	UpdateScrollBar();
	handled = true;
	break;
    case sclRepeatEvent:
	ScrollLines(e->data.sclRepeat.newValue - e->data.sclRepeat.value,
		    false);
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID) {
	case SendButton:
	    /* Check that we have a user */
	    if (prefs->lastUser == 0) {
		FrmAlert(NoUserAlert);
	    } else {
		/* Get pointer to current user record */
		r = GetUserRecord(prefs->userRecord,
				  prefs->lastUser - 1);
		if (r == NULL) {
		    FrmAlert(NoUserAlert);
		    break;
		}
		if (prefs->lastAccess == 0)
		    a = NULL;
		else
		    a = GetAccessRecord(r->accessRecord,
					prefs->lastAccess - 1);
		/* Get pointers to the various things to read from */
		frm = FrmGetFormPtr(MainForm);
		sfield = FrmGetObjectPtr(frm,
				   FrmGetObjectIndex(frm, SubjectField));
		efield = FrmGetObjectPtr(frm,
				     FrmGetObjectIndex(frm, EntryField));
		mfield = FrmGetObjectPtr(frm,
				      FrmGetObjectIndex(frm, MoodField));
		mlist = FrmGetObjectPtr(frm,
					FrmGetObjectIndex(frm, MoodList));

		/* And post! */
		if (!Post(r->user, a == NULL ? NULL : a->user, r->passwd,
			  FldGetTextPtr(sfield),
			  FldGetTextPtr(efield),
			  LstGetSelection(mlist),
			  FldGetTextPtr(mfield),
			  prefs->connectType)) {
		    FrmAlert(NetErrAlert);
		} else {
		    FrmAlert(NetOkayAlert);
		    FldSetText(sfield, NULL, 0, 0);
		    FldSetText(efield, NULL, 0, 0);
		    FldSetText(mfield, NULL, 0, 0);
		    LstSetSelection(mlist, 0);
		    trigger = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							   MoodTrigger));
		    CtlSetLabel(trigger, LstGetSelectionText(mlist, 0));
		    FrmHideObject(frm, FrmGetObjectIndex(frm, MoodField));
		    FrmUpdateForm(UserForm, frmRedrawUpdateCode);
		}
	    }
	    handled = true;
	    break;
	}
	break;

    default:
	break;
    }

    return handled;
}

static Boolean AboutFormHandleEvent(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	handled = true;
	break;

    case ctlSelectEvent:
	handled = true;
	FrmReturnToForm(0);
	break;

    default:
	break;
    }

    return handled;
}

static Boolean ConnectionFormHandleEvent(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    ListPtr clist;
    ControlPtr trigger;

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	clist = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ConnectionList));
	trigger = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
						     ConnectionTrigger));
	if (prefs->connectType == Undecided)
	    LstSetSelection(clist, 0);
	else
	    LstSetSelection(clist, prefs->connectType - 1);
	CtlSetLabel(trigger, LstGetSelectionText(clist,
						 LstGetSelection(clist)));
	FrmDrawForm(frm);
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	handled = true;
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID) {
	case OKButton:
	    handled = true;
	    frm = FrmGetActiveForm();
	    clist = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							ConnectionList));
	    prefs->connectType = LstGetSelection(clist) + 1;
	    SavePrefs(prefs);
	    FrmReturnToForm(0);
	    break;
	case CancelButton:
	    handled = true;
	    FrmReturnToForm(0);
	    break;
	}
	break;

    default:
	break;
    }

    return handled;
}

static void UpdateUserList(ListPtr ulist)
{
    UInt16 i, j, k, l;
    UInt16 lines;
    UserRecord *r;
    AccessRecord *b;
    static MemHandle h = NULL;
    static Char *a = NULL;
    Char *p;

    /* Calculate big string length */
    lines = 0;
    for (i = 0, j = 0; i < prefs->userCount; i++) {
	r = GetUserRecord(prefs->userRecord, i);
	j += StrLen(r->name) + 1;
	if (r->accessRecord != NULL) {
	    for (k = 0; k < r->accessCount; k++) {
		b = GetAccessRecord(r->accessRecord, k);
		j += StrLen(b->user) + 10;
		lines++;
	    }
	}
	lines++;
    }
    /* Make big string */
    if (prefs->userCount == 0) {
	LstSetListChoices(ulist, NULL, 0);
    } else {
	/* Prepare user/access map */
	if (usermap != NULL)
	    MemPtrFree(usermap);
	if (accessmap != NULL)
	    MemPtrFree(accessmap);
	usermap = MemPtrNew(sizeof(UInt16) * lines);
	accessmap = MemPtrNew(sizeof(UInt16) * lines);
	if (a != NULL)
	    MemPtrFree(a);
	if (h != NULL)
	    MemHandleFree(h);
	a = MemPtrNew(j);
	for (i = 0, p = a, l = 0; i < prefs->userCount; i++) {
	    r = GetUserRecord(prefs->userRecord, i);
	    MemMove(p, r->name, StrLen(r->name));
	    p += StrLen(r->name);
	    p[0] = '\0';
	    p++;
	    usermap[l] = i;
	    accessmap[l++] = -1;
	    if (r->accessRecord != NULL) {
		for (k = 0; k < r->accessCount; k++) {
		    MemMove(p, "    - as ", 9);
		    p += 9;
		    b = GetAccessRecord(r->accessRecord, k);
		    MemMove(p, b->user, StrLen(b->user));
		    p += StrLen(b->user);
		    p[0] = '\0';
		    p++;
		    usermap[l] = i;
		    accessmap[l++] = k;
		}
	    }
	}
	/* Make list from strings */
	h = SysFormPointerArrayToStrings(a, lines);
	LstSetListChoices(ulist, MemHandleLock(h), lines);
    }
}

static Boolean UserFormHandleEvent(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    ListPtr ulist;

    UserRecord *r, *s;
    int i;

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();

	FrmDrawForm(frm);
	handled = true;
	break;

    case winExitEvent:
	frm = FrmGetActiveForm();
	ulist = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, UserList));
	UpdateUserList(ulist);
	LstSetSelection(ulist, prefs->lastUser - 1);
	FrmUpdateForm(UserForm, frmRedrawUpdateCode);
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	handled = true;
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID) {
	case NewButton:
	    if (prefs->connectType == Undecided)
		FrmAlert(NoConnectionAlert);
	    else
		FrmPopupForm(NewUserForm);
	    break;
	case UpdateButton:
	    if (prefs->connectType == Undecided) {
		FrmAlert(NoConnectionAlert);
		break;
	    }
	    frm = FrmGetActiveForm();
	    ulist = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							   UserList));
	    if (LstGetNumberOfItems(ulist) == 0)
		break;
	    r = GetUserRecord(prefs->userRecord,
			      usermap[LstGetSelection(ulist)]);
	    if (!Login(r->user, r->passwd, prefs, false, false,
		       prefs->connectType))
		FrmAlert(BadLoginAlert);
	    else {
		UpdateUserList(ulist);
		FrmUpdateForm(UserForm, frmRedrawUpdateCode);
	    }
	    break;
	case DeleteButton:
	    frm = FrmGetActiveForm();
	    ulist = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							   UserList));
	    i = LstGetNumberOfItems(ulist);
	    if (i == 0)
		break;
	    if (!FrmAlert(DeleteAlert)) {
		i = LstGetSelection(ulist);
		if (i == 0) {
		    r = DelUserRecord(prefs->userRecord);
		    prefs->userRecord = r;
		    prefs->userCount--;
		    UpdateUserList(ulist);
		} else {
		    r = GetUserRecord(prefs->userRecord, usermap[i] - 1);
		    s = GetUserRecord(prefs->userRecord, usermap[i]);
		    s = DelUserRecord(s);
		    r->next = s;
		    prefs->userCount--;
		    UpdateUserList(ulist);
		}
		FrmUpdateForm(UserForm, frmRedrawUpdateCode);
	    }
	    break;
	default:
	    frm = FrmGetActiveForm();
	    ulist = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							   UserList));
	    if (LstGetNumberOfItems(ulist) > 0) {
		prefs->lastUser = usermap[LstGetSelection(ulist)] + 1;
		prefs->lastAccess = accessmap[LstGetSelection(ulist)] + 1;
	    }
	    SavePrefs(prefs);
	    handled = 1;
	    FrmReturnToForm(0);
	    break;
	}
	break;

    default:
	break;
    }

    return handled;
}

static Boolean NewUserFormHandleEvent(EventPtr e)
{
    Boolean handled = false;
    FormPtr frm;
    FieldPtr ufield;
    Char *user;
    FieldPtr pfield;
    Char *passwd;

    switch (e->eType) {
    case frmOpenEvent:
	frm = FrmGetActiveForm();
	FrmDrawForm(frm);
	handled = true;
	break;

    case menuEvent:
	MenuEraseStatus(NULL);

	handled = true;
	break;

    case ctlSelectEvent:
	switch (e->data.ctlSelect.controlID) {
	case UOKButton:
	    frm = FrmGetActiveForm();
	    ufield = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							    UnameField));
	    pfield = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm,
							    PasswdField));
	    user = FldGetTextPtr(ufield);
	    passwd = FldGetTextPtr(pfield);
	    if (StrLen(user) == 0 || StrLen(passwd) == 0) {
		FrmAlert(MoreCharsAlert);
		break;
	    }
	    if (!Login(user, passwd, prefs, false, true, prefs->connectType)) {
		FrmAlert(BadLoginAlert);
		break;
	    }
	    if (prefs->lastUser == 0)
		prefs->lastUser = 1;
	    SavePrefs(prefs);
	    handled = true;
	    FrmReturnToForm(0);
	    break;
	case CancelButton:
	    handled = true;
	    FrmReturnToForm(0);
	    break;
	}
	break;

    default:
	break;
    }

    return handled;
}


static Boolean ApplicationHandleEvent(EventPtr e)
{
    FormPtr frm;
    UInt16 formId;
    Boolean handled = false;

    if (e->eType == frmLoadEvent) {
	formId = e->data.frmLoad.formID;
	frm = FrmInitForm(formId);
	FrmSetActiveForm(frm);

	switch (formId) {
	case MainForm:
	    FrmSetEventHandler(frm, MainFormHandleEvent);
	    break;
	case AboutForm:
	    FrmSetEventHandler(frm, AboutFormHandleEvent);
	    break;
	case UserForm:
	    FrmSetEventHandler(frm, UserFormHandleEvent);
	    break;
	case NewUserForm:
	    FrmSetEventHandler(frm, NewUserFormHandleEvent);
	    break;
	case ConnectionForm:
	    FrmSetEventHandler(frm, ConnectionFormHandleEvent);
	    break;
	}
	handled = true;
    }
    return handled;
}

/* Get preferences, open (or create) app database */
static UInt16 StartApplication(void)
{
    prefs = MemPtrNew(sizeof(LJPrefs));
    FrmGotoForm(MainForm);
    return 0;
}

/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
    FrmSaveAllForms();
    FrmCloseAllForms();
    MemPtrFree(prefs);
}

/* The main event loop */
static void EventLoop(void)
{
    UInt16 err;
    EventType e;

    do {
	EvtGetEvent(&e, evtWaitForever);
	if (!SysHandleEvent(&e))
	    if (!MenuHandleEvent(NULL, &e, &err))
		if (!ApplicationHandleEvent(&e))
		    FrmDispatchEvent(&e);
    } while (e.eType != appStopEvent);
}

/* Main entry point */
UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
    UInt16 err;

    if (cmd == sysAppLaunchCmdNormalLaunch) {

	err = StartApplication();
	if (err)
	    return err;

	EventLoop();
	StopApplication();

    } else {
	return sysErrParamErr;
    }

    return 0;
}
